///////////////////////////////////////////////////////////////////////////////
// This file is part of "MMV - MegaloManic Venture", which is a demo which took
// part in Ohjelmointiputka's "Demokilpailu 2005" in January 2005.
//
// Copyleft Matti "Deewiant" Niemenmaa, 2004-2005.
//
// Filename: beginning.cc
//           All the code from the very start to the beginning of the year 2005
//           run-through
//
// File created: 2005-01-02
///////////////////////////////////////////////////////////////////////////////

#include "header.hh"

/******************************************************************************\
Function introText()

Responsible for showing all the text at the start, from "DEEWIANT" to the title
of the demo.
\******************************************************************************/
void introText()
{
	double aika;
	int green = 0, blue = 0, green2 = 0, blue2 = 0, red = 0;

	glfwSetTime(0);

	do {
		aika = glfwGetTime();
		glViewport(0, 0, 1024, 768);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();

		glColor3ub(0, green, blue);
		// only check for one of blue, green since they're both the same
		if (blue < 128 && aika < 4) {
			blue = ++green;
		}

		if (aika < 6) {
			kirjoita(fonts[proFontBig], 280, 448, "DEEWIANT");
			kirjoita(fonts[proFont],    320, 448, "\npresents");
		}
		if (aika >= 1 && aika < 6) {
			if (green2 < 128 && aika < 4) {
				glColor3ub(0, green2, blue2);
				green2 = ++blue2;
			}
			kirjoita(fonts[proFont], 360, 448, "\n\nin association with");
		}

		if (aika >= 3 && blue > 0 && aika < 6)
			green = blue = blue2 = --green2;

		if (aika >= 5 && green == 0 && aika <= 11) {
			glColor3ub(red, 0, blue2);
			if (aika <= 10 && blue2 < 144) {
				++blue2;
				if (red < 42)
					++red;
			}
			kirjoita(fonts[proFontBig], 392, 604, "GURU");
			kirjoita(fonts[proFontBig], 360, 16,  "POSSE");

			// GP logo
			pushScreenCoordinateMatrix();
			glColor3ub(blue2, blue2, blue2);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glEnable(GL_BLEND);
			glEnable(GL_LINE_SMOOTH);
			glLineWidth(3);
			glBegin(GL_LINE_LOOP);
				// top
				glVertex2f( 512, 576   );
				// top-left
				glVertex2f( 320, 380.4 );
				// bottom-right
				glVertex2f( 704, 64    );
				// top
				glVertex2f( 512, 576   );
				// bottom-left
				glVertex2f( 320, 64    );
				// top-right
				glVertex2f( 704, 380.4 );
				// top
				glVertex2f( 512, 576   );
			glEnd();
			glLineWidth(1);
			glDisable(GL_LINE_SMOOTH);
			glDisable(GL_BLEND);

			popProjectionMatrix();
		}

		if (aika >= 9 && blue2 > 0) {
			--blue2;
			if (red > 0)
				--red;
		}

		if (aika >= 11) {
			glColor3ub(255, 0, 72);
			kirjoita(fonts[proFontLarge], 384, 536, "MMV -");
			kirjoita(fonts[proFontBig], 224, 384, "MegaloManic\nVenture");
		}

		if (aika >= 12) {
			glColor3ub(blue, blue, blue);
			if (blue < 224)
				++blue;
			kirjoita(fonts[proFontSmall], 0, 168, "Originally released for the\nDemo Competition at Ohjelmointiputka\nheld on January 20, 2005");
		}

		glPopMatrix();

		glfwSwapBuffers();

	} while (aika < 15 && glfwGetWindowParam(GLFW_OPENED));

	return;
}

#include "Vector.hh"
#include "Ray.hh"

/******************************************************************************\
Struct Plane

A simple struct containing the two pieces of information needed to define a
plane - its distance from the origin and a vector perpendicular to it.
\******************************************************************************/
struct Plane {
	Vector _position;
	Vector _normal;
};

int const ballNum = 42;
Vector velocity[ballNum];
Vector position[ballNum];
Vector oldPos  [ballNum];
Plane theFloor, wall;
double const Time = 0.6;

// radius & diameter of one ball
int const r = 8;
int const d = 2 * r;

/******************************************************************************\
Function planeVectorIsect()

Determines whether a plane and a vector intersect.

Inputs:
	plane      The plane.
	position   The position of the vector.
	direction  Ditto direction.
	colDist    A memory location where to place the distance
	           to the collision point.
	pNormal    Ditto normal of the plane.

Outputs:
	true or false, depending on whether an intersection takes place.
\******************************************************************************/
bool planeVectorIsect(Plane const & plane, Vector const & position, Vector const & direction, double & colDist, Vector & pNormal)
{
	double scalarProduct = direction.dot(plane._normal);
	double collDist;

	// reasonably close to zero - determine if the vector is parallel to the plane
	if (scalarProduct < 1.0e-8 && scalarProduct > -1.0e-8)
		return false;

	collDist = (plane._normal.dot(plane._position - position)) / scalarProduct;

	if (collDist < -1.0e-8)
		return false;

	pNormal = plane._normal;
	colDist = collDist;
	return true;
}

/******************************************************************************\
Function ballBallIsect()

Determines whether two spheres intersect.

Inputs:
	point    A memory location where the collision point will be stored.
	time     Ditto for the time of collision.
	time2    How often we will check.
	ballOne  The array index of the first sphere.
	ballTwo  Ditto second.
\******************************************************************************/

bool ballBallIsect(Vector & point, double & time, double time2, int & ballOne, int & ballTwo)
{
	Vector relativeVec;
	Ray rays;
	double myTime = 0, add = time2/150, timeDummy = 10000;
	Vector posi;

	for (int i = 0; i < ballNum - 1; ++i) {
		for (int j = i + 1; j < ballNum; ++j) {
			// find distance
			relativeVec = velocity[i] - velocity[j];
			rays = Ray(oldPos[i], Vector::unit(relativeVec));
			myTime = 0;

			// if the balls don't touch (distance between is > their radii)
			if ((rays.dist(oldPos[j])) > d) continue;
			// else (not needed since continue jumps out of loop if true)
			// find exact intersection point
			while (myTime < time2) {
				myTime += add;
				posi = oldPos[i] + (relativeVec * myTime);
				if (posi.dist(oldPos[j]) <= d) {
					point = posi;
					if (timeDummy > (myTime - add))
						timeDummy = myTime - add;
					ballOne = i;
					ballTwo = j;
					break;
				}
			}
		}
	}

	if (timeDummy != 10000) {
		time = timeDummy;
		return true;
	}

	return false;
}

bool wallExists = false, noWalls = false;

/******************************************************************************\
Function do2005Physics()

The whole physics engine used with the balls in drop2005() - collision
detection, velocity calculation, etc.

It's a pretty slow function that needs optimisation, but I didn't have the
time... my 1.2 GHz Duron skipped a few frames at one point, but that's not so
bad, considering most people nowadays have faster CPUs anyways.
\******************************************************************************/

/* having had to start this whole thing over once, I was pretty far behind schedule
** so I decided to just copy the code from NeHe's OpenGL lesson #30
** without really taking a look at what's actually happening there
** and trusting him to have made it so it'll work =)
** hence the less-decipherable code here
** btw, it's amazing how well the guy can fuck up his indentation, inserting
** random spaces or tabs everywhere... I figure he was working with a tab stop
** of 4 (it's the only value that looked legible to me), but even with that
** the code is pretty ugly... ah well, the code works, that suffices
*/
void do2005Physics()
{
	double rt,rt2,rt4,lamda=10000;
	Vector norm,uveloc;
	Vector normal,point,time;
	double RestTime,BallTime;
	Vector Pos2;
	int BallNr=0,BallColNr1,BallColNr2;
	Vector Nc;

	RestTime=Time;

	//While timestep not over
	while (RestTime > 1.0e-8) {
		lamda=10000;   //initialize to very large value

		//For all the balls find closest intersection between balls and plane
		for (int i=0;i<ballNum;i++) {
			//compute new position and distance
			oldPos[i]=position[i];
			Vector::unit(velocity[i],uveloc);
			position[i]=position[i]+velocity[i]*RestTime;
			rt2=oldPos[i].dist(position[i]);

			if (!noWalls) {
				//Test if collision occured between ball and plane
				if (planeVectorIsect(theFloor,oldPos[i],uveloc,rt,norm)) {
					//Find intersection time
					rt4=rt*RestTime/rt2;

					//if smaller than the one already stored replace and in timestep
					if (rt4<=lamda) {
						if (rt4<=RestTime+1.0e-8)
							if (! ((rt<=1.0e-8)&&(uveloc.dot(norm)>1.0e-8)) ) {
								normal=norm;
								point=oldPos[i]+uveloc*rt;
								lamda=rt4;
								BallNr=i;
							}
					}
				}
				if (wallExists) {
					if (planeVectorIsect(wall,oldPos[i],uveloc,rt,norm)) {
						//Find intersection time
						rt4=rt*RestTime/rt2;

						//if smaller than the one already stored replace and in timestep
						if (rt4<=lamda) {
							if (rt4<=RestTime+1.0e-8)
								if (! ((rt<=1.0e-8)&&(uveloc.dot(norm)>1.0e-8)) ) {
									normal=norm;
									point=oldPos[i]+uveloc*rt;
									lamda=rt4;
									BallNr=i;
								}
						}
					}
				}
			}
		}

		//After all balls were teste with plane test for collision
		//between them and replace if collision time smaller
		if (ballBallIsect(Pos2,BallTime,RestTime,BallColNr1,BallColNr2)) {
			if ( (lamda==10000) || (lamda>BallTime) ) {
				RestTime=RestTime-BallTime;

				Vector pb1,pb2,xaxis,U1x,U1y,U2x,U2y,V1x,V1y,V2x,V2y;
				double a,b;

				pb1=oldPos[BallColNr1]+velocity[BallColNr1]*BallTime;
				pb2=oldPos[BallColNr2]+velocity[BallColNr2]*BallTime;
				xaxis=(pb2-pb1).unit();

				a=xaxis.dot(velocity[BallColNr1]);
				U1x=xaxis*a;
				U1y=velocity[BallColNr1]-U1x;

				xaxis=(pb1-pb2).unit();
				b=xaxis.dot(velocity[BallColNr2]);
				U2x=xaxis*b;
				U2y=velocity[BallColNr2]-U2x;

				V1x=(U1x+U2x-(U1x-U2x))*0.5;
				V2x=(U1x+U2x-(U2x-U1x))*0.5;
				V1y=U1y;
				V2y=U2y;

				for (int j=0;j<ballNum;j++)
					position[j]=oldPos[j]+velocity[j]*BallTime;

				velocity[BallColNr1]=V1x+V1y;
				velocity[BallColNr2]=V2x+V2y;

				continue;
			}
		}

		//End of tests
		//If test occured move simulation for the correct timestep
		//and compute response for the colliding ball
		if (lamda!=10000) {
			RestTime-=lamda;

			for (int j=0;j<ballNum;j++)
				position[j]=oldPos[j]+velocity[j]*lamda;

			rt2=velocity[BallNr].mag();
			velocity[BallNr].unit();
			velocity[BallNr]=Vector::unit( (normal*(2*normal.dot(-velocity[BallNr]))) + velocity[BallNr] );
			velocity[BallNr]=velocity[BallNr]*rt2;

		} else RestTime=0;
	}

	return;
}

/******************************************************************************\
Function drop2005()

The bit where the coloured balls that form "2005" move around and bounce.
\******************************************************************************/
void drop2005()
{
	for (int i = 0; i < ballNum; ++i)
		velocity[i] = Vector(0, -7.00, 0);

	// d is the MAXIMUM distance between balls which causes infinite collision

	unsigned short const sx = 256, sy = 1280, z = r - 1;
	double const q = d + 0.00001;
	// 2
	position[0] = Vector(sx     , sy-q   , z);
	position[1] = Vector(sx+q   , sy     , z);
	position[2] = Vector(sx+2*q , sy     , z);
	position[3] = Vector(sx+3*q , sy-q   , z);
	position[4] = Vector(sx+3*q , sy-2*q , z);
	position[5] = Vector(sx+2*q , sy-3*q , z);
	position[6] = Vector(sx     , sy-4*q , z);
	position[7] = Vector(sx+q   , sy-4*q , z);
	position[8] = Vector(sx+2*q , sy-4*q , z);
	position[9] = Vector(sx+3*q , sy-4*q , z);
	// 0
	position[10] = Vector(sx+q  , sy-6*q , z);
	position[11] = Vector(sx+2*q, sy-6*q , z);
	position[12] = Vector(sx+3*q, sy-7*q , z);
	position[13] = Vector(sx+3*q, sy-8*q , z);
	position[14] = Vector(sx+3*q, sy-9*q , z);
	position[15] = Vector(sx+2*q, sy-10*q, z);
	position[16] = Vector(sx+q  , sy-10*q, z);
	position[17] = Vector(sx    , sy-9*q , z);
	position[18] = Vector(sx    , sy-8*q , z);
	position[19] = Vector(sx    , sy-7*q , z);
	// 0
	position[20] = Vector(sx+q  , sy-12*q, z);
	position[21] = Vector(sx+2*q, sy-12*q, z);
	position[22] = Vector(sx+3*q, sy-13*q, z);
	position[23] = Vector(sx+3*q, sy-14*q, z);
	position[24] = Vector(sx+3*q, sy-15*q, z);
	position[25] = Vector(sx+2*q, sy-16*q, z);
	position[26] = Vector(sx+q  , sy-16*q, z);
	position[27] = Vector(sx    , sy-15*q, z);
	position[28] = Vector(sx    , sy-14*q, z);
	position[29] = Vector(sx    , sy-13*q, z);
	// 5
	position[30] = Vector(sx+3*q, sy-18*q, z);
	position[31] = Vector(sx+2*q, sy-18*q, z);
	position[32] = Vector(sx+q  , sy-18*q, z);
	position[33] = Vector(sx    , sy-18*q, z);
	position[34] = Vector(sx    , sy-19*q, z);
	position[35] = Vector(sx    , sy-20*q, z);
	position[36] = Vector(sx+q  , sy-20*q, z);
	position[37] = Vector(sx+2*q, sy-20*q, z);
	position[38] = Vector(sx+3*q, sy-21*q, z);
	position[39] = Vector(sx+2*q, sy-22*q, z);
	position[40] = Vector(sx+q  , sy-22*q, z);
	position[41] = Vector(sx    , sy-22*q, z);

	theFloor._position = Vector(0, 0, 0);
	theFloor._normal   = Vector(0, 1, 0);

	double aika, aika2 = 10000;
	int white = 255, red = 255, out = 0;
	bool one=false, two=false, three=false, four=false, five=false, six=false, seven=false;

	GLUquadricObj * quadric = gluNewQuadric();

	glfwSetTime(0);

	do {
		aika = glfwGetTime();
		do2005Physics();
		glViewport(0, 0, 1024, 768);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glLoadIdentity();

		// draw balls
		for (int i = 0; i < ballNum; ++i) {
			glPushMatrix();
			pushScreenCoordinateMatrix();
			glTranslated(position[i].X(), position[i].Y(), position[i].Z());

			if (i <= 9)
				glColor3ub(0, 255, 0);
			else if (i >= 10 && i <= 19)
				glColor3ub(255, 0, 0);
			else if (i >= 20 && i <= 29)
				glColor3ub(0, 255, 255);
			else
				glColor3ub(192, 128, 192);

			gluSphere(quadric, r, 24, 24);
			popProjectionMatrix();
			glPopMatrix();
		}

		if (position[41].Y() <= 192 && position[41].X() == sx && !one) {
			one = true;
			for (int i = 0; i < ballNum; ++i)
				velocity[i] = Vector(0, 0, 0);

			wallExists = true;

			wall._position = Vector(1020, 384, z);
			wall._normal   = Vector(-1, 0, 0);

			aika2 = glfwGetTime();
		}

		// 1 second later
		if (glfwGetTime() - aika2 >= 0.9 && glfwGetTime() - aika2 <= 1.1 && !two) {
			two = true;
			for (int i = 0; i <= 29; ++i)
				velocity[i] = Vector(0, -5, 0);
			for (int i = 30; i <= 41; ++i)
				velocity[i] = Vector(15, 0, 0);
		}

		if (position[25].Y() <= 192 && position[25].X() <= (sx + 3 * q) && !three) {
			three = true;
			for (int i = 20; i <= 29; ++i)
				velocity[i] = Vector(15, 0, 0);
		}

		if (position[15].Y() <= 192 && position[15].Y() <= (sx + 3 * q) && !four) {
			four = true;
			for (int i = 10; i <= 19; ++i)
				velocity[i] = Vector(15, 0, 0);
		}

		if (position[6].Y() <= 224 && !five) {
			five = true;
			for (int i = 0; i <= ballNum; ++i)
				velocity[i] = Vector(0, 0, 0);

			// strange bug... position of 0th ball suddenly becomes {0, 0, 0}
			if (position[0].Z() != z) {
				position[0] = Vector(256, 272.8, z);
			}

			six = true;

			aika = glfwGetTime();
		}

		if (six) {

			glfwSetTime(0);

			if (white > 0)
				--white;

			glColor3ub(white, white, white);
			kirjoita(fonts[proFont], 768, 384, "Oops!!");

			if (white == 0 && aika < 10 && red > 0) {
				--red;

				glColor3ub(red, 0, 0);
				kirjoita(fonts[proFontBig], 16, 576, "Oh, forget it!");

				if (!seven) {
					seven = true;
					for (int i = 0; i < ballNum; ++i)
						velocity[i] = Vector(80 - i, -i, 0);
				}
			}

			for (int i = 0; i < ballNum; ++i) {
				if (position[i].X() < -10 || position[i].Y() > 778)
					++out;
				else out = 0;
				if (out == 42) goto ballsEnd; // two-level break
			}
		}

		glfwSwapBuffers();

	} while (aika < 16 && glfwGetWindowParam(GLFW_OPENED));

	ballsEnd:

	signed short const s2x = 384, s2y = -256;

	position[0] = Vector(s2x    ,   s2y-q  , z);
	position[1] = Vector(s2x+q  ,   s2y    , z);
	position[2] = Vector(s2x+2*q,   s2y    , z);
	position[3] = Vector(s2x+3*q,   s2y-q  , z);
	position[4] = Vector(s2x+3*q,   s2y-2*q, z);
	position[5] = Vector(s2x+2*q,   s2y-3*q, z);
	position[6] = Vector(s2x    ,   s2y-4*q, z);
	position[7] = Vector(s2x+q  ,   s2y-4*q, z);
	position[8] = Vector(s2x+2*q,   s2y-4*q, z);
	position[9] = Vector(s2x+3*q,   s2y-4*q, z);
	// 0
	position[10] = Vector(s2x+6*q,  s2y,     z);
	position[11] = Vector(s2x+7*q,  s2y,     z);
	position[12] = Vector(s2x+8*q,  s2y-q,   z);
	position[13] = Vector(s2x+8*q,  s2y-2*q, z);
	position[14] = Vector(s2x+8*q,  s2y-3*q, z);
	position[15] = Vector(s2x+7*q,  s2y-4*q, z);
	position[16] = Vector(s2x+6*q,  s2y-4*q, z);
	position[17] = Vector(s2x+5*q,  s2y-3*q, z);
	position[18] = Vector(s2x+5*q,  s2y-2*q, z);
	position[19] = Vector(s2x+5*q,  s2y-q,   z);
	// 0
	position[20] = Vector(s2x+11*q, s2y,     z);
	position[21] = Vector(s2x+12*q, s2y,     z);
	position[22] = Vector(s2x+13*q, s2y-q,   z);
	position[23] = Vector(s2x+13*q, s2y-2*q, z);
	position[24] = Vector(s2x+13*q, s2y-3*q, z);
	position[25] = Vector(s2x+12*q, s2y-4*q, z);
	position[26] = Vector(s2x+11*q, s2y-4*q, z);
	position[27] = Vector(s2x+10*q, s2y-3*q, z);
	position[28] = Vector(s2x+10*q, s2y-2*q, z);
	position[29] = Vector(s2x+10*q, s2y-q,   z);
	// 5
	position[30] = Vector(s2x+18*q, s2y,     z);
	position[31] = Vector(s2x+17*q, s2y,     z);
	position[32] = Vector(s2x+16*q, s2y,     z);
	position[33] = Vector(s2x+15*q, s2y,     z);
	position[34] = Vector(s2x+15*q, s2y-q,   z);
	position[35] = Vector(s2x+15*q, s2y-2*q, z);
	position[36] = Vector(s2x+16*q, s2y-2*q, z);
	position[37] = Vector(s2x+17*q, s2y-2*q, z);
	position[38] = Vector(s2x+18*q, s2y-3*q, z);
	position[39] = Vector(s2x+17*q, s2y-4*q, z);
	position[40] = Vector(s2x+16*q, s2y-4*q, z);
	position[41] = Vector(s2x+15*q, s2y-4*q, z);

	do {
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glColor3f(1, 1, 1);

		for (int i = 0; i < ballNum; ++i) {
			glPushMatrix();
			glTranslated(position[i].X(), position[i].Y(), position[i].Z());
			gluSphere(quadric, r, 16, 16);
			glPopMatrix();
		}

		if (position[25].Y() < 64)
			for (int i = 0; i < ballNum; ++i)
				position[i] += Vector(0, 4, 0);
		else { glfwSwapBuffers(); break; }

		glfwSwapBuffers();

	} while (glfwGetWindowParam(GLFW_OPENED));

	gluDeleteQuadric(quadric);

	return;
}
