#ifndef HEADER_HH
#define HEADER_HH

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <GL/glfw.h>
#include "audiere.h"

// evil, I know, but I had only an hour or so of time when adding Audiere...
// so I couldn't be bothered to add the audiere::s everywhere.
// Upon reflection, it might have taken me longer to write this comment. Whatever.
using namespace audiere;

extern "C" {
#include <jpeglib.h>
}

// Degrees to Radians, Radians to Degrees, and half of Pi, respectively.
namespace GEO {
	double const DE2RA      = 0.01745329252;
	double const RA2DE      = 57.2957795129;
	double const PIOVER2    = 1.570796326795;
}

/******************************************************************************\
Struct fontData

Holds all the data we'll need to use any font.
\******************************************************************************/
struct fontData {
	double         height;
	unsigned int * textures;
	unsigned int   startId;

	void makeFont(char const *, unsigned int);

	void clean();
};

/******************************************************************************\
Struct jpegData

The data for a JPEG. Used only when loading and creating a texture from a JPEG.
\******************************************************************************/
struct jpegData {
	int rowSpan, sizeX, sizeY;
	unsigned char * data;
};

/******************************************************************************\
Struct smallVec

"Small" vectors, in that they are a) 2D vectors and B) capable of moving only
parallel to the Y axis.
\******************************************************************************/
struct smallVec {
	double x, y, yVel;
};

extern struct jpegData     * imageData;
extern bool                  counter;
extern unsigned short        year, month, day;
extern fontData              fonts[];
extern unsigned int          textures[];
extern smallVec              bouncyBalls[];
extern GLfloat       const   allLightIntensity[], atmosLightIntensity[];
extern GLfloat               atmosLightPosition[];
extern double                rotation[];

enum {
	proFontTiny,
	proFontSmall,
	proFont,
	proFontBig,
	proFontLarge
};

enum {
	earth,
	norway,
	imp
};

// main.cc
void initialize         (   );
void finish             (   );
void onResize(int, int);

// beginning.cc
void introText          (   );
void drop2005           (   );

// middle.cc
void newYearFireworks   (   );
void Ireland            (   );
void chineseNewYear     (   );
void eclipse            (int);
void hhgttg             (   );
void NorwayIndependence (   );
void Doom               (   );
extern bool takeTime;
extern bool drawEarth;

// end.cc
void someText           (   );
void credits            (   );
void Deewiant           (   );
void markdark           (   );
void theRealEnd         (   );

// text.cc
void kirjoita(fontData const &, double, double, char const *, ...);

// general.cc
void pushScreenCoordinateMatrix (    void    );
void popProjectionMatrix        (    void    );
void doLower                    (    void    );
void setTop                     (    void    );
void doEarth                    (    void    );
void counterGoto                (short, short);

// debugging stuff
void report(char const *);
extern int debugCnt;

#endif
