///////////////////////////////////////////////////////////////////////////////
// This file is part of "MMV - MegaloManic Venture", which is a demo which took
// part in Ohjelmointiputka's "Demokilpailu 2005" in January 2005.
//
// Copyleft Matti "Deewiant" Niemenmaa, 2004-2005.
//
// Filename: main.cc
//           Entry and exit point of the program.
//
// File created: 2004-12-31
///////////////////////////////////////////////////////////////////////////////

#include "header.hh"
#include "minifmod.h"

// #define DEBUG

#ifdef DEBUG
int debugCnt = 0;
int const windowType = GLFW_WINDOW;
void GLFWCALL onResize(int, int);
#else
int const windowType = GLFW_FULLSCREEN;
#endif
void GLFWCALL onKey(int, int);

// for music
FMUSIC_MODULE * module;

// for fonts
int const fontNum = 5;
fontData  fonts[fontNum];

// for textures
struct jpegData * EarthData,
		* flagData,
		* impData;
int const         texNum = 3;
unsigned int      textures[texNum];

// for the counter
bool               counter;
unsigned short int year = 2005, month = 0, day = 1;

// for the physics bit - bouncing balls
smallVec bouncyBalls[10];

// for lighting
GLfloat const allLightIntensity  [] = { 0.75, 0.75, 0.75, 1 };
GLfloat const atmosLightIntensity[] = { 3, 3,  3, 1 };
GLfloat       atmosLightPosition [] = { 0, 0, -1, 0 };

// rotation starts with 80 so that we don't have to spin on January 1st...
double rotation[2] = { 0, 80 };

// This was to be the only function that does not have a comment associated with it.
// It didn't end up that way.
int main()
{
	initialize();
	report("Demo initialised.");

	introText();
	report("Intro text written.");
	drop2005();
	report("Dropped numbers.");
	for (int i = 0; i < 10; ++i) {
		bouncyBalls[i].x = 892 + (i * 8.001);
		bouncyBalls[i].y = i * 8 + 88;
	}
	report("Balls set bouncing.");
	// From this point on, every function draws the bottom frame every frame by calling doLower()
	report("Lower frame is in use.");
	report("Top frame is ready.");

	report("Earth is in sight.");
	counter = true;
	report("Counter started.");

	newYearFireworks();
	report("Fireworks fired.");
	counterGoto(1, 20);
	Ireland();
	report("Internationalised.");
	counterGoto(2, 9);
	chineseNewYear();
	counterGoto(4, 8);
	eclipse(0);
	report("Eclipse 1 shown.");
	counterGoto(5, 6);
	hhgttg();
	report("Earth demolished.");
	counterGoto(6, 7);
	NorwayIndependence();
	report("Norwegians happy.");
	counterGoto(10, 3);
	eclipse(1);
	report("Eclipse 2 shown.");
	counterGoto(12, 10);
	Doom();
	report("Doom is 12! Jubilation!");
	someText();
	report("End oneliners shown.");
	credits();
	report("Credits displayed.");
	Deewiant();
	report("Deewiant orated.");
	markdark();
	report("markdark orated also.");
	theRealEnd();
	report("As of now, all is well.");

	finish();

	// should never be reached
	report("Something went very wrong!");
	return 0;
}

/******************************************************************************\
Function fileopen()
Function fileclose()
Function fileread()
Function fileseek()
Function filetell()

Callback functions for FMOD.
\******************************************************************************/
unsigned int fileopen(char * name)
{
	return (reinterpret_cast<unsigned int>(fopen(name, "rb")));
}
void fileclose(unsigned int handle)
{
	fclose(reinterpret_cast<FILE *>(handle));
	return;
}
int fileread(void *buffer, int size, unsigned int handle)
{
	return fread(buffer, 1, size, reinterpret_cast<FILE *>(handle));
}
void fileseek(unsigned int handle, int pos, signed char mode)
{
	fseek(reinterpret_cast<FILE *>(handle), pos, mode);
	return;
}
int filetell(unsigned int handle)
{
	return ftell(reinterpret_cast<FILE *>(handle));
}

/******************************************************************************\
Function initialize()

Initialises the graphics and sound by way of GLFW and fMod, and reads in any
files used - fonts, textures, sounds.

The JPEG loading and processing parts should really be in a function of their
own, but I couldn't get it to work (gave me runtime errors, probably bad pointer
handling) in the 20-or-so minutes time that I had when adding the second
texture, and I was pretty short on time anyway, so I just did it like this,
quickly...
\******************************************************************************/
void initialize()
{
	struct jpeg_decompress_struct deStruct;
	struct jpeg_error_mgr jerr;
	EarthData = new jpegData;

	deStruct.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&deStruct);

	FILE * file;
	if ((file = std::fopen("Data/Earth.jpeg", "rb")) == 0) {
		std::fprintf(stderr, "Failed to load Data/Earth.jpeg texture.");
		std::exit(true);
	}
	jpeg_stdio_src(&deStruct, file);

	jpeg_read_header(&deStruct, TRUE);
	jpeg_start_decompress(&deStruct);

	EarthData->rowSpan = deStruct.image_width * deStruct.num_components;
	EarthData->sizeX   = deStruct.image_width;
	EarthData->sizeY   = deStruct.image_height;

	EarthData->data = new unsigned char[EarthData->rowSpan * EarthData->sizeY];

	unsigned char ** rowPtr = new unsigned char * [EarthData->sizeY];
	for (int i = 0; i < EarthData->sizeY; i++)
		rowPtr[i] = &(EarthData->data[i*EarthData->rowSpan]);

	int rowsRead = 0;

	// output_scanline is the library's own state variable
	while (deStruct.output_scanline < deStruct.output_height) {
		// read in the current row of pixels and increase the rowsRead count
		rowsRead += jpeg_read_scanlines(&deStruct, &rowPtr[rowsRead], deStruct.output_height - rowsRead);
	}

	// Delete the temporary row pointers
	delete [] rowPtr;

	jpeg_finish_decompress(&deStruct);
	jpeg_destroy_decompress(&deStruct);
	std::fclose(file);

// Earth loaded, moving on to Norway

	struct jpeg_decompress_struct deStruct2;
	struct jpeg_error_mgr jerr2;
	flagData = new jpegData;

	deStruct2.err = jpeg_std_error(&jerr2);
	jpeg_create_decompress(&deStruct2);

	FILE * file2;
	if ((file2 = std::fopen("Data/Norway.jpg", "rb")) == 0) {
		std::fprintf(stderr, "Failed to load Data/Norway.jpg texture.");
		std::exit(true);
	}
	jpeg_stdio_src(&deStruct2, file2);

	jpeg_read_header(&deStruct2, TRUE);
	jpeg_start_decompress(&deStruct2);

	flagData->rowSpan = deStruct2.image_width * deStruct2.num_components;
	flagData->sizeX   = deStruct2.image_width;
	flagData->sizeY   = deStruct2.image_height;

	flagData->data = new unsigned char[flagData->rowSpan * flagData->sizeY];

	unsigned char ** rowPtr2 = new unsigned char * [flagData->sizeY];
	for (int i = 0; i < flagData->sizeY; i++)
		rowPtr2[i] = &(flagData->data[i*flagData->rowSpan]);

	rowsRead = 0;

	// output_scanline is the library's own state variable
	while (deStruct2.output_scanline < deStruct2.output_height) {
		// read in the current row of pixels and increase the rowsRead count
		rowsRead += jpeg_read_scanlines(&deStruct2, &rowPtr2[rowsRead], deStruct2.output_height - rowsRead);
	}

	// Delete the temporary row pointers
	delete [] rowPtr2;

	jpeg_finish_decompress(&deStruct2);
	jpeg_destroy_decompress(&deStruct2);
	std::fclose(file2);

// Norway loaded, moving on to imp

	struct jpeg_decompress_struct deStruct3;
	struct jpeg_error_mgr jerr3;
	impData = new jpegData;

	deStruct3.err = jpeg_std_error(&jerr3);
	jpeg_create_decompress(&deStruct3);

	FILE * file3;
	if ((file3 = std::fopen("Data/imp.jfif", "rb")) == 0) {
		std::fprintf(stderr, "Failed to load Data/imp.jfif texture.");
		std::exit(true);
	}
	jpeg_stdio_src(&deStruct3, file3);

	jpeg_read_header(&deStruct3, TRUE);
	jpeg_start_decompress(&deStruct3);

	impData->rowSpan = deStruct3.image_width * deStruct3.num_components;
	impData->sizeX   = deStruct3.image_width;
	impData->sizeY   = deStruct3.image_height;

	impData->data = new unsigned char[impData->rowSpan * impData->sizeY];

	unsigned char ** rowPtr3 = new unsigned char * [impData->sizeY];
	for (int i = 0; i < impData->sizeY; i++)
		rowPtr3[i] = &(impData->data[i*impData->rowSpan]);

	rowsRead = 0;

	// output_scanline is the library's own state variable
	while (deStruct3.output_scanline < deStruct3.output_height) {
		// read in the current row of pixels and increase the rowsRead count
		rowsRead += jpeg_read_scanlines(&deStruct3, &rowPtr3[rowsRead], deStruct3.output_height - rowsRead);
	}

	// Delete the temporary row pointers
	delete [] rowPtr3;

	jpeg_finish_decompress(&deStruct3);
	jpeg_destroy_decompress(&deStruct3);
	std::fclose(file3);

// imp loaded, moving on to rest of program

	// initialise FMOD
	FSOUND_File_SetCallbacks(fileopen, fileclose, fileread, fileseek, filetell);

	if (!FSOUND_Init(44100, 0)) {
		std::fprintf(stderr, "Could not initialise sound (FMOD).");
		std::exit(true);
	}

	if (!(module = FMUSIC_LoadSong("Data/2005mix.xm", NULL))) {
		FSOUND_Close();
		std::fprintf(stderr, "Could not load the music file Data/2005mix.xm.");
		std::exit(true);
	}

	// initialise GLFW
	if (!glfwInit()) {
		std::fprintf(stderr, "Could not initialise graphics (GLFW).");
		std::exit(true);
	}

	if (!glfwOpenWindow(1024, 768, 8, 8, 8, 8, 24, 0, windowType)) {
		glfwTerminate();
		std::fprintf(stderr, "Could not open a window.");
		std::exit(true);
	}

	glfwSetWindowTitle("MMV - MegaloManic Venture");
	glfwEnable(GLFW_STICKY_KEYS);

	// JPEG processing...
	glGenTextures(1, &textures[earth]);
	glBindTexture(GL_TEXTURE_2D, textures[earth]);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, EarthData->sizeX, EarthData->sizeY, GL_RGB, GL_UNSIGNED_BYTE, EarthData->data);
	if (EarthData) {
		if (EarthData->data)
			delete EarthData->data;
		delete EarthData;
	}
	glGenTextures(1, &textures[norway]);
	glBindTexture(GL_TEXTURE_2D, textures[norway]);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, flagData->sizeX, flagData->sizeY, GL_RGB, GL_UNSIGNED_BYTE, flagData->data);
	if (flagData) {
		if (flagData->data)
			delete flagData->data;
		delete flagData;
	}
	glGenTextures(1, &textures[imp]);
	glBindTexture(GL_TEXTURE_2D, textures[imp]);
	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, impData->sizeX, impData->sizeY, GL_RGB, GL_UNSIGNED_BYTE, impData->data);
	if (impData) {
		if (impData->data)
			delete impData->data;
		delete impData;
	}
	// done processing JPEGs.

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glfwSetKeyCallback(onKey);
#ifdef DEBUG
	glfwSetWindowSizeCallback(onResize);
#endif

	fonts[proFontTiny].  makeFont("Data/ProFontWindows.ttf", 12);
	fonts[proFontSmall]. makeFont("Data/ProFontWindows.ttf", 20);
	fonts[proFont].      makeFont("Data/ProFontWindows.ttf", 48);
	fonts[proFontBig].   makeFont("Data/ProFontWindows.ttf", 96);
	fonts[proFontLarge]. makeFont("Data/ProFontWindows.ttf", 144);

	glLightfv(GL_LIGHT1, GL_AMBIENT, allLightIntensity);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, atmosLightIntensity);

	FMUSIC_PlaySong(module);

	pushScreenCoordinateMatrix();

	return;
}

/******************************************************************************\
Function finish()

Terminates GLFW and fMod, and cleans up any leftover resources (hopefully).
\******************************************************************************/
void finish()
{
	for (int i = 0; i < fontNum; ++i)
		fonts[i].clean();

	FMUSIC_StopSong(module);
	FMUSIC_FreeSong(module);
	FSOUND_Close();

	glfwTerminate();

	report("Demo terminated.");

	std::exit(false);

	return;
}

/******************************************************************************\
Function onKey()

Dictates the action to be taken when a key is pressed. In this case, it only
does something if the key was Esc and it was released - it quits the demo.

Inputs:
	key      The id of the pressed key
	action   The id of the action (pressed or released)
\******************************************************************************/
void GLFWCALL onKey(int key, int action)
{
	if (action == GLFW_RELEASE && key == GLFW_KEY_ESC)
		finish();
	else return;
}

// debugging functions
#ifdef DEBUG
void report(char const * const asdf) {std::printf(asdf); std::printf("\t%2d\n", debugCnt); ++debugCnt; return;}
void GLFWCALL onResize(int width, int height) { glfwSetWindowSize(1024, 768); }
#else
void report(char const * const asdf) {}
#endif
