///////////////////////////////////////////////////////////////////////////////
// This file is part of "MMV - MegaloManic Venture", which is a demo which took
// part in Ohjelmointiputka's "Demokilpailu 2005" in January 2005.
//
// Copyleft Matti "Deewiant" Niemenmaa, 2004-2005.
// well, actually, this one is mostly Copyleft NeHe
//
// Filename: text.cc
//           Code for drawing text on the screen using Freetype
//
// File created: 2005-01-02
///////////////////////////////////////////////////////////////////////////////

#include <string>
#include <vector>
#include <ft2build.h>
#include <freetype/freetype.h>
#include <freetype/ftglyph.h>
#include <freetype/ftoutln.h>
#include <freetype/fttrigon.h>

#include "header.hh"

/******************************************************************************\
Function nextP2()

Gets the first power of 2 which is >= the given number.

Inputs:
	num      The number from which we start looking

Outputs:
	p2       The next power of 2 to num
\******************************************************************************/
int nextP2(int num)
{
	int p2 = 1;
	while (p2 < num) p2 <<= 1;
	return p2;
}

/******************************************************************************\
Function makeDisplayList()

Creates a display list used by a font for whichever character we pass it.

Inputs:
	face       The font we're using
	kar        The character we want the display list for
	startId    The beginning id for the textures array
	textures   Pointer to an array where we'll put the lists
\******************************************************************************/
void makeDisplayList(FT_Face face, unsigned char kar, unsigned int startId, unsigned int * textures)
{
	// load the glyph
	if (FT_Load_Glyph(face, FT_Get_Char_Index(face, kar), FT_LOAD_DEFAULT)) {
		std::perror("Failed to load a glyph for a character.");
		std::exit(true);
	}

	// put the glyph into a glyph object
	FT_Glyph glyph;
	if (FT_Get_Glyph(face->glyph, &glyph)) {
		std::perror("Failed to put a glyph in a glyph.");
		std::exit(true);
	}

	// turn the glyph into a bitmap
	FT_Glyph_To_Bitmap(&glyph, ft_render_mode_normal, 0, 1);
	FT_BitmapGlyph bitmap_glyph = reinterpret_cast<FT_BitmapGlyph>(glyph);

	// get the widths of the bitmap data, needed to create the textures
	int width =  nextP2( bitmap_glyph -> bitmap . width );
	int height = nextP2( bitmap_glyph -> bitmap . rows  );

	// memory for texture data
	GLubyte * expandedData = new GLubyte[2 * width * height];

	// fill in the data for the bitmaps (luminosity and alpha)
	// (very arcane code - put 0 if we are in the "padding zone" which was
	//  created when we made the width/height the next power of 2)
	for (int i = 0; i < height; ++i) {
		for (int j = 0; j < width; ++j) {
			  expandedData[2 * (j + (i * width))]
			= expandedData[2 * (j + (i * width)) + 1]
			= (j >= bitmap_glyph->bitmap.width || i >= bitmap_glyph->bitmap.rows)
			  ? 0 : bitmap_glyph->bitmap.buffer[j + (bitmap_glyph->bitmap.width * i)];
		}
	}

	// setup texture
	glBindTexture   (GL_TEXTURE_2D, kar[textures]);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	// create texture
	glTexImage2D    (GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_BYTE, expandedData);

	// we have a texture - don't need the data to make it any more
	delete [] expandedData;

	// make the display list itself
	glNewList(startId + kar, GL_COMPILE);

	glBindTexture(GL_TEXTURE_2D, textures[kar]);

	glPushMatrix();

	// move a bit so there's space between the characters
	glTranslatef(bitmap_glyph->left, 0, 0);

	// move down a bit if character extends below line, like y or g
	glTranslatef(0, (bitmap_glyph->top - bitmap_glyph->bitmap.rows), 0);

	// account for the fact that many chars have padding space
	// figure out what's used by char itself, store that in x/y
	double x =   static_cast<float>(bitmap_glyph->bitmap.width)
	           / static_cast<float>(width);
	double y =   static_cast<float>(bitmap_glyph->bitmap.rows)
	           / static_cast<float>(height);

	// draw a quad and map the font texture to it
	// using a quad makes life easier when trying to move/rotate/whatever it
	glBegin(GL_QUADS);
		glTexCoord2d(0, 0); glVertex2f(0, bitmap_glyph->bitmap.rows);
		glTexCoord2d(0, y); glVertex2f(0, 0);
		glTexCoord2d(x, y); glVertex2f(bitmap_glyph->bitmap.width, 0);
		glTexCoord2d(x, 0); glVertex2f(bitmap_glyph->bitmap.width,
		                               bitmap_glyph->bitmap.rows);
	glEnd();
	glPopMatrix();
	glTranslatef(face->glyph->advance.x >> 6, 0, 0);

	// list is done
	glEndList();

	return;
}

/******************************************************************************\
Function fontData::makeFont()

Loads a font from a file and makes it ready for use.

Inputs:
	name     The name of the file from where we take the font
	height   The wanted height of the font
\******************************************************************************/
void fontData::makeFont(char const * name, unsigned int height)
{
	// memory for textures
	textures = new unsigned int[128];

	this->height = height;

	// initialize a Freetype library
	FT_Library library;
	if (FT_Init_FreeType(&library)) {
		std::perror("Failed to initialise a Freetype library.");
		std::exit(true);
	}

	// initialise a Freetype font face
	FT_Face face;
	if (FT_New_Face(library, name, 0, &face)) {
		std::perror("Failed to load the font.");
		std::exit(true);
	}

	// Freetype font sizes are in 1/64th of pixels, so use height * 64
	FT_Set_Char_Size(face, height << 6, height << 6, 96, 96);

	// resources for textures and display lists
	startId = glGenLists(128);
	glGenTextures(128, textures);

	// make display lists
	for (unsigned char i = 0; i < 128; ++i)
		makeDisplayList(face, i, startId, textures);

	// display lists made, these no longer needed
	FT_Done_Face(face);
	FT_Done_FreeType(library);

	return;
}

/******************************************************************************\
Function fontData::clean()

Frees any memory which would be left behind when finishing with a font.
\******************************************************************************/
void fontData::clean()
{
	glDeleteLists(startId, 128);
	glDeleteTextures(128, textures);
	delete [] textures;
}

/******************************************************************************\
Function kirjoita()

Writes text to the screen.

Inputs:
	theFont    The font we'll be using
	x, y       Coordinates, where to write (lower left corner of 1st letter)
	tekst      The text we'll be writing
\******************************************************************************/
void kirjoita(fontData const & theFont, double x, double y, char const * tekst, ...)
{
	// we want a window coordinate system
	pushScreenCoordinateMatrix();

	unsigned int font = theFont.startId;

	// increase the height a bit, put some space between lines
	double height = theFont.height / 0.63;

	// I know, not smart, but suffices for MMV
	char text[1536];

	// process arguments...
	va_list args;
	if (tekst == 0)
		*text = 0;
	else {
		va_start(args, tekst);
		vsprintf(text, tekst, args);
		va_end(args);
	}

	// split the text into a set of lines
	char const * startLine = text;
	std::vector<std::string> lines;
	char const * c = text;
	for (;*c; ++c) {
		if (*c == '\n') {
			std::string line;
			for (char const * d = startLine; d < c; ++d)
				line.append(1, *d);
			lines.push_back(line);
			startLine = c + 1;
		}
	}
	if (startLine) {
		std::string line;
		for (char const * d = startLine; d < c; d++)
			line.append(1, *d);
		lines.push_back(line);
	}

	// set some attributes...
	glPushAttrib(GL_LIST_BIT | GL_CURRENT_BIT | GL_ENABLE_BIT | GL_TRANSFORM_BIT);
	glMatrixMode(GL_MODELVIEW);
	glDisable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glListBase(font);

	float modelViewMatrix[16];
	glGetFloatv(GL_MODELVIEW_MATRIX, static_cast<GLfloat*>(modelViewMatrix));

	// do actual text display
	// for each new line, reset modelview matrix
	for (unsigned int i = 0; i < lines.size(); ++i) {
		glPushMatrix();
		glLoadIdentity();
		glTranslatef(x, y - (height * i), 0);
		glMultMatrixf(modelViewMatrix);

		glCallLists(lines[i].length(), GL_UNSIGNED_BYTE, lines[i].c_str());

		glPopMatrix();
	}

	glPopAttrib();

	popProjectionMatrix();
}
